//
//	Invert.cpp
//

#define  WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <tchar.h>

#define WND_CLASS		_T("INVERT")
#define APP_TITLE		_T("Invert")

#define INIT_WIDTH		320
#define INIT_HEIGHT		240

#define MIN_WIDTH		10
#define MIN_HEIGHT		10

#ifndef max
#define max(a, b)		((a) > (b) ? (a) : (b))
#endif

LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	static HDC hDC = NULL, hDC_Mem = NULL, hDC_DT = NULL;
	static HBITMAP hBmp_Prev = NULL;
	static UINT nMouseFlag = 0;
	static POINT ptStart = {0, 0}, ptPos = {0, 0};
	switch (uMsg) {
	case WM_CREATE:
		{
			HWND hWnd_DT = GetDesktopWindow();
			RECT rc;
			GetWindowRect(hWnd_DT, &rc);
			LONG cx = rc.right - rc.left, cy = rc.bottom - rc.top;
			hDC_DT = GetDC(hWnd_DT);
			hDC_Mem = CreateCompatibleDC(hDC_DT);
			hBmp_Prev = (HBITMAP)SelectObject(hDC_Mem, CreateCompatibleBitmap(hDC_DT, cx, cy));
			BitBlt(hDC_Mem, 0, 0, cx, cy, hDC_DT, 0, 0, SRCCOPY);
			hDC = GetDC(hWnd);
		}
		break;
	case WM_LBUTTONDOWN:
		if (!nMouseFlag) {
			nMouseFlag = 1;
			ptStart.x = (LONG)(short)LOWORD(lParam);
			ptStart.y = (LONG)(short)HIWORD(lParam);
			GetCursorPos(&ptPos);
			SetCapture(hWnd);
		}
		break;
	case WM_RBUTTONDOWN:
		if (!nMouseFlag) {
			nMouseFlag = 2;
			RECT rc;
			GetWindowRect(hWnd, &rc);
			ptStart.x = (LONG)(short)LOWORD(lParam) - (rc.right - rc.left);
			ptStart.y = (LONG)(short)HIWORD(lParam) - (rc.bottom - rc.top);
			SetCapture(hWnd);
		}
		break;
	case WM_MOUSEMOVE:
		switch (nMouseFlag) {
		case 1:
			{
				RECT rc;
				GetWindowRect(hWnd, &rc);
				LONG cxWnd = rc.right - rc.left;
				LONG cyWnd = rc.bottom - rc.top;
				POINT pt;
				GetCursorPos(&pt);
				LONG cxMove = pt.x - ptPos.x;
				LONG cyMove = pt.y - ptPos.y;
				if (cxMove || cyMove) {
					if (cxMove < 0) {
						BitBlt(hDC_Mem, rc.left + cxMove, rc.top, -cxMove, cyWnd, hDC_DT, rc.left + cxMove, rc.top, SRCCOPY);
					} else if (cxMove > 0) {
						BitBlt(hDC_Mem, rc.right, rc.top, cxMove, cyWnd, hDC_DT, rc.right, rc.top, SRCCOPY);
					}
					if (cyMove < 0) {
						BitBlt(hDC_Mem, rc.left, rc.top + cyMove, cxWnd, -cyMove, hDC_DT, rc.left, rc.top + cyMove, SRCCOPY);
					} else if (cyMove > 0) {
						BitBlt(hDC_Mem, rc.left, rc.bottom, cxWnd, cyMove, hDC_DT, rc.left, rc.bottom, SRCCOPY);
					}
					LONG x = pt.x - ptStart.x;
					LONG y = pt.y - ptStart.y;
					SetWindowPos(hWnd, NULL, x, y, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
					BitBlt(hDC, 0, 0, cxWnd, cyWnd, hDC_Mem, x, y, NOTSRCCOPY);
					ptPos = pt;
				}
			}
			break;
		case 2:
			{
				RECT rc;
				GetWindowRect(hWnd, &rc);
				LONG cxWnd = rc.right - rc.left;
				LONG cyWnd = rc.bottom - rc.top;
				LONG cx = (LONG)(short)LOWORD(lParam) - ptStart.x;
				LONG cy = (LONG)(short)HIWORD(lParam) - ptStart.y;
				if (cx < MIN_WIDTH ) cx = MIN_WIDTH ;
				if (cy < MIN_HEIGHT) cy = MIN_HEIGHT;
				if (cx > cxWnd) {
					BitBlt(hDC_Mem, rc.right, rc.top, cx - cxWnd, max(cy, cyWnd), hDC_DT, rc.right, rc.top, SRCCOPY);
				}
				if (cy > cyWnd) {
					BitBlt(hDC_Mem, rc.left, rc.bottom, max(cx, cxWnd), cy - cyWnd, hDC_DT, rc.left, rc.bottom, SRCCOPY);
				}
				SetWindowPos(hWnd, NULL, 0, 0, cx, cy, SWP_NOMOVE | SWP_NOZORDER);
				BitBlt(hDC, 0, 0, cx, cy, hDC_Mem, rc.left, rc.top, NOTSRCCOPY);
			}
			break;
		}
		break;
	case WM_LBUTTONUP:
		if (nMouseFlag == 1) {
			nMouseFlag = 0;
			ReleaseCapture();
		}
		break;
	case WM_RBUTTONUP:
		if (nMouseFlag == 2) {
			nMouseFlag = 0;
			ReleaseCapture();
		}
		break;
	case WM_PAINT:
		{
			RECT rc;
			GetWindowRect(hWnd, &rc);
			PAINTSTRUCT ps;
			BeginPaint(hWnd, &ps);
			BitBlt(hDC, ps.rcPaint.left, ps.rcPaint.top, ps.rcPaint.right - ps.rcPaint.left, ps.rcPaint.bottom - ps.rcPaint.top,
				hDC_Mem, rc.left + ps.rcPaint.left, rc.top + ps.rcPaint.top, NOTSRCCOPY);
			EndPaint(hWnd, &ps);
		}
		break;
	case WM_KEYDOWN:
		switch (wParam) {
		case VK_ESCAPE:
			SendMessage(hWnd, WM_CLOSE, 0, 0);
			break;
		}
		break;
	case WM_DESTROY:
		{
			DeleteObject(SelectObject(hDC_Mem, hBmp_Prev));
			DeleteDC(hDC_Mem);
			ReleaseDC(hWnd, hDC);
			ReleaseDC(GetDesktopWindow(), hDC_DT);
		}
		PostQuitMessage(0);
		break;
	default:
		return DefWindowProc(hWnd, uMsg, wParam, lParam);
	}
	return 0;
}

int APIENTRY _tWinMain(HINSTANCE hInstance, HINSTANCE, LPTSTR, int nCmdShow) {
	WNDCLASSEX wcx		= { sizeof(WNDCLASSEX), };
	wcx.lpfnWndProc		= WndProc;
	wcx.hInstance		= hInstance;
	wcx.hCursor			= LoadCursor(NULL, IDC_ARROW);
	wcx.hbrBackground	= (HBRUSH)(COLOR_WINDOW + 1);
	wcx.lpszClassName	= WND_CLASS;
	RegisterClassEx(&wcx);
	HWND hWnd = CreateWindowEx(WS_EX_TOPMOST, wcx.lpszClassName, APP_TITLE, WS_POPUP,
					CW_USEDEFAULT, 0, INIT_WIDTH, INIT_HEIGHT, NULL, NULL, wcx.hInstance, NULL);
	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);
	MSG msg;
	while (GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return int(msg.wParam);
}
